/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimal working examples (MWE) to build upon and 
	 modify to gain an understanding of the API

	 Extensions to the Vulkan Graphics Coffee Break Series
	   
/*
/****************************************************************/


#version 420

// Vertex attributes
layout (location = 0) in vec4 inPos;
layout (location = 1) in vec2 inUV;
layout (location = 2) in vec3 inColor;
layout (location = 3) in vec3 inNormal;
layout (location = 4) in uint inMatId;

// Instance attributes
layout (location = 5) in vec3 instancePos;
layout (location = 6) in vec3 instanceRot;


layout (binding = 0) uniform UBO 
{
	mat4  model;
	mat4  view;
	mat4  projection;
	vec4  lightPos;
} ubo;

layout (location = 0) out vec3  outNormal;
layout (location = 1) out vec3  outColor;
layout (location = 2) out vec3  outEyePos;
layout (location = 3) out vec3  outLightVec;
layout (location = 4) out vec2  outUV;
layout (location = 5) out uint  outMatId;

out gl_PerVertex
{
	vec4 gl_Position;
};

void main() 
{
	// build our rotation matrix
	mat4 yrot;
	// only rotate around y for testing (x,z todo)
	float s = sin(instanceRot.y);
	float c = cos(instanceRot.y);
	yrot[0] = vec4(c,   0.0, s,   0.0);
	yrot[1] = vec4(0.0, 1.0, 0.0, 0.0);
	yrot[2] = vec4(-s,  0.0, c,   0.0);
	yrot[3] = vec4(0.0, 0.0, 0.0, 1.0);

	vec4 pos    = vec4(inPos.xyz + instancePos, 1);
	outNormal   = vec3( ubo.model * vec4(inNormal,0) );
	outColor    = inColor;
	gl_Position = ubo.projection * ubo.view * (ubo.model * pos);
	outEyePos   = vec3(ubo.view * ubo.model * pos);
	outLightVec = -normalize(ubo.lightPos.xyz - outEyePos);


	outUV    = inUV;
	outMatId = inMatId;
}
