/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/
/*
	Texture wrapper class - load tga image - we 
	load the tga pixels in (32bit tga) and 
	generate the Vulkan image/view - ready for use in our
	graphical scene.

*/


#pragma once

#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


#include "helpers.h"




class VkTGATexture
{
public:
	VkSampler				sampler			    = NULL;
	VkImage					image			    = NULL;
	VkImageView				imageView		    = NULL;
	VkDeviceMemory			imageMemory		    = { 0 };
	VkDescriptorImageInfo	descriptorImageInfo;
	int						width			    = 0;
	int						height			    = 0;

	struct TGAFILE
	{
		unsigned char  imageTypeCode;
		short int      imageWidth;
		short int      imageHeight;
		unsigned char  bitCount;
		unsigned char *imageData;
	};


	VkTGATexture(const char*	  tgaFileName,
				 VkDevice		  device,
			     VkPhysicalDevice physicalDevice)
	{
		VkSamplerCreateInfo samplerCreateInfo = {};
		samplerCreateInfo.sType             = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
		samplerCreateInfo.magFilter         = VK_FILTER_LINEAR;
		samplerCreateInfo.minFilter         = VK_FILTER_LINEAR;
		samplerCreateInfo.mipmapMode        = VK_SAMPLER_MIPMAP_MODE_LINEAR;
		samplerCreateInfo.addressModeU      = VK_SAMPLER_ADDRESS_MODE_REPEAT;
		samplerCreateInfo.addressModeV      = VK_SAMPLER_ADDRESS_MODE_REPEAT;
		samplerCreateInfo.addressModeW      = VK_SAMPLER_ADDRESS_MODE_REPEAT;
		samplerCreateInfo.mipLodBias       = 0;
		samplerCreateInfo.anisotropyEnable = VK_FALSE;
		samplerCreateInfo.minLod           = 0;
		samplerCreateInfo.maxLod           = 5;
		samplerCreateInfo.borderColor      = VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
		samplerCreateInfo.unnormalizedCoordinates = VK_FALSE;

    
		VkResult result = 
		vkCreateSampler( device, 
						 &samplerCreateInfo, 
						 NULL, 
						 &sampler );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create sampler." );

		TGAFILE tgaFile;
		bool err =
		LoadTGAFile( tgaFileName, &tgaFile );
		DBG_ASSERT(err);

		width  = tgaFile.imageWidth;
		height = tgaFile.imageHeight;
		

		DBG_ASSERT(tgaFile.bitCount == 32 );
		void* imageData = (void*)tgaFile.imageData;

		
		VkImageCreateInfo textureCreateInfo = {};
		textureCreateInfo.sType         = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		textureCreateInfo.imageType     = VK_IMAGE_TYPE_2D;
		textureCreateInfo.format        = VK_FORMAT_R8G8B8A8_UNORM;;
		textureCreateInfo.extent.width  = width; 
		textureCreateInfo.extent.height = height; 
		textureCreateInfo.extent.depth  = 1; 
		textureCreateInfo.mipLevels     = 1;
		textureCreateInfo.arrayLayers   = 1;
		textureCreateInfo.samples       = VK_SAMPLE_COUNT_1_BIT;
		textureCreateInfo.tiling        = VK_IMAGE_TILING_LINEAR;
		textureCreateInfo.usage         = VK_IMAGE_USAGE_SAMPLED_BIT;
		textureCreateInfo.sharingMode   = VK_SHARING_MODE_EXCLUSIVE;
		textureCreateInfo.initialLayout = VK_IMAGE_LAYOUT_PREINITIALIZED;

		result = 
		vkCreateImage( device, 
					   &textureCreateInfo, 
					   NULL, 
					   &image );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create texture image." );

		// allocate and bind image memory:
		VkMemoryRequirements textureMemoryRequirements = {};
		vkGetImageMemoryRequirements( device, 
									  image, 
									  &textureMemoryRequirements );

		VkMemoryAllocateInfo textureImageAllocateInfo = {};
		textureImageAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		textureImageAllocateInfo.allocationSize = textureMemoryRequirements.size;

		// read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

		uint32_t textureMemoryTypeBits = textureMemoryRequirements.memoryTypeBits;
		VkMemoryPropertyFlags tDesiredMemoryFlags = VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT;
		for( uint32_t i = 0; i < 32; ++i ) 
		{
			VkMemoryType memoryType = memoryProperties.memoryTypes[i];
			if( textureMemoryTypeBits & 1 ) 
			{
				if( ( memoryType.propertyFlags & tDesiredMemoryFlags ) == tDesiredMemoryFlags ) 
				{
					textureImageAllocateInfo.memoryTypeIndex = i;
					break;
				}
			}
			textureMemoryTypeBits = textureMemoryTypeBits >> 1;
		}

		result = vkAllocateMemory( device, &textureImageAllocateInfo, NULL, &imageMemory );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to allocate device memory." );

		result = vkBindImageMemory( device, image, imageMemory, 0 );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to bind image memory." );

		// copy our image content:
		void *imageMapped;
		result = vkMapMemory( device, imageMemory, 0, VK_WHOLE_SIZE, 0, &imageMapped );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to map uniform buffer memory." );
    
		memcpy( imageMapped, imageData, sizeof(DWORD) * width * height  );

		VkMappedMemoryRange memoryRange = {};
		memoryRange.sType  = VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE;
		memoryRange.memory = imageMemory;
		memoryRange.offset = 0;
		memoryRange.size   = VK_WHOLE_SIZE;
		vkFlushMappedMemoryRanges( device, 1, &memoryRange );
    
		vkUnmapMemory( device, imageMemory );

	
		free(imageData);
		imageData = NULL;

		// create and map the texture view
		VkImageViewCreateInfo textureImageViewCreateInfo = {};
		textureImageViewCreateInfo.sType    = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		textureImageViewCreateInfo.image    = image;
		textureImageViewCreateInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
		textureImageViewCreateInfo.format   = VK_FORMAT_R8G8B8A8_UNORM;//VK_FORMAT_R32G32B32_SFLOAT;
		VkComponentMapping j = { VK_COMPONENT_SWIZZLE_R, 
								 VK_COMPONENT_SWIZZLE_G, 
								 VK_COMPONENT_SWIZZLE_B, 
								 VK_COMPONENT_SWIZZLE_A };
		textureImageViewCreateInfo.components                      = j; 
		textureImageViewCreateInfo.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_COLOR_BIT;
		textureImageViewCreateInfo.subresourceRange.baseMipLevel   = 0;
		textureImageViewCreateInfo.subresourceRange.levelCount     = 1;
		textureImageViewCreateInfo.subresourceRange.baseArrayLayer = 0;
		textureImageViewCreateInfo.subresourceRange.layerCount     = 1;

		result = 
		vkCreateImageView( device, 
						   &textureImageViewCreateInfo, 
						   NULL, 
						   &imageView );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create image view." );

		// Fill descriptor image info that can be used for setting up descriptor sets
		descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
		descriptorImageInfo.imageView   = imageView;
		descriptorImageInfo.sampler     = sampler;

	}

	



	bool LoadTGAFile(const char *filename, TGAFILE *tgaFile)
	{
		FILE *filePtr;
		unsigned char ucharBad;
		short int sintBad;
		long imageSize;
		int colorMode;
		unsigned char colorSwap;

		// Open the TGA file.
		filePtr = fopen(filename, "rb");
		if (filePtr == NULL)
		{
			return false;
		}

		// Read the two first bytes we don't need.
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);

		// Which type of image gets stored in imageTypeCode.
		fread(&tgaFile->imageTypeCode, sizeof(unsigned char), 1, filePtr);

		// For our purposes, the type code should be 2 (uncompressed RGB image)
		// or 3 (uncompressed black-and-white images).
		if (tgaFile->imageTypeCode != 2 && tgaFile->imageTypeCode != 3)
		{
			fclose(filePtr);
			return false;
		}

		// Read 13 bytes of data we don't need.
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);
		fread(&sintBad, sizeof(short int), 1, filePtr);

		// Read the image's width and height.
		fread(&tgaFile->imageWidth, sizeof(short int), 1, filePtr);
		fread(&tgaFile->imageHeight, sizeof(short int), 1, filePtr);

		// Read the bit depth.
		fread(&tgaFile->bitCount, sizeof(unsigned char), 1, filePtr);

		// Read one byte of data we don't need.
		fread(&ucharBad, sizeof(unsigned char), 1, filePtr);

		// Color mode -> 3 = BGR, 4 = BGRA.
		colorMode = tgaFile->bitCount / 8;
		imageSize = tgaFile->imageWidth * tgaFile->imageHeight * colorMode;

		// Allocate memory for the image data.
		tgaFile->imageData = (unsigned char*)malloc(sizeof(unsigned char)*imageSize);

		// Read the image data.
		fread(tgaFile->imageData, sizeof(unsigned char), imageSize, filePtr);

		// Change from BGR to RGB so OpenGL can read the image data.
		for (int imageIdx = 0; imageIdx < imageSize; imageIdx += colorMode)
		{
			colorSwap = tgaFile->imageData[imageIdx];
			tgaFile->imageData[imageIdx] = tgaFile->imageData[imageIdx + 2];
			tgaFile->imageData[imageIdx + 2] = colorSwap;
		}

		fclose(filePtr);
		return true;
	}

};